; **********************************************************
; Einstein Experimental Fat Loader V1.0(C) Phil Simmons 2007
; **********************************************************
; This is all my own work but I'd like to credit 
; the CHAOS-OS Ultimate Boot Loader for IntelX86
; from which much help & inspiration was derived
; and label names were shamelessly stolen :)

; I use the TASM32 assembler from Squak Valley Software
; which is 32 bit command line macro assembler - it also
; supports many other microprocessors (like the 6502)

#Define EQU .EQU
#Define DEFB .DB
#Define DB   .DB
#Define DW .DW
#Define DEFW .DW
#Define DEFM .TEXT
#Define DM .TEXT
#Define DS .FILL
#Define ORG .ORG


;  Einstein MOS Calls

MCAL			EQU $CF 	        ; RST 08h MOS calls
ZINIT			EQU $9A		        ; MOS Re-entry
ZKEYIN			EQU $9C		        ; MOS Wait for key press & return char
ZRBLK                   EQU $A4                 ; MOS Read Block from disk
ZCRLF			EQU $A6		        ; MOS Output A Newline
ZHMDSC			EQU $AF		        ; MOS Reset Drive to T0/S0
ZIMULT			EQU $CE		        ; MOS multiplies  two 16 bit values
ZIN80			EQU $D3		        ; MOS Detect 40/80 card and set as appropriate
ZSYSRS			EQU $BE		        ; MOS Reset system
ZPRM 			EQU $CF		        ; MOS print mesg ends with 80h
CR			EQU $0D
LF			EQU $0A
HI			EQU $80
SP			EQU $20

BREAK                   EQU $FF

; Special values for FAT entries


FAT16_BAD_CLUS	        EQU $FFFD               ; These are my own values not MS
FAT16_RESERVED	        EQU $FFFE
FAT16_LAST_CLUS	        EQU $FFFF

; Directory entry attributes

FAT_ATTR_READ_ONLY	EQU 01                  ; From MS but not used here
FAT_ATTR_HIDDEN	        EQU 02
FAT_ATTR_SYSTEM	        EQU 04
FAT_ATTR_VOLUME_ID	EQU 08
FAT_ATTR_DIRECTORY	EQU 10

	                ORG 100H

; BootSector Fat Structure for BDOS

; ALL einstein boot disks use the first 3 words - functions as labelled

LoadStart:              DW 0100H
LoadEnd:                DW 0300H
ExecAddr:               DW Entry                ; Jump to the OS loader

Entry:                  JR BootStrap

; *** Boot Parameter Block (BPB) - used to calculate position of root,fat and data  ***

DeviceId:               DB 0                    ; Media type + drive num, FDD=0;HDD=80h eg.FDD 1 =01
MaxRootEntries:	        DW 128		        ; For Now
BytesPerSector:         DW 512	                ; As for all drives to be used
SecPerTrk:		DW 10                   ; Einstein = 0-9
SectorsPerCluster:	DW 2		        ; n * 512 bytes / 1k Allocation unit for now
RsvdSectCnt:	        DW 1		        ; MS FAT 12/16 standard
TotalSect16:	        DW 400		        ; 40 Trks * 10 Trk/Sec, S/S  for now
TotalSect32:	        DW 0		        ; Zero if TotalSect16 defined
SectorShift:            DW 0                    ; 4 giga sector partition
FatSize16:		DW 2		        ; Sectors ( 512 bytes ) per FAT
FatSize32               DW 0                    ; Zero if Fat16
NumHeads:		DW 1                    ; reserved for CHS
NumCylinders:           DW 0                    ; reserved for CHS
BS_Reserved1:	        DB 0		        ; not used for now

; ****** End of Boot Parameter Block *******


; Loader Code starts here

BootStrap: 	        LD (OldSP),SP           ; store stack pointer
			LD SP,BootStack	        ; Avoid using the MOS stack
                        LD A,(DeviceId)	        ; From BPB
                        AND 3                   ; Mask off drive num > 3

; Re-entry here with drive number in A

DriveRetry:             LD (BootDrive),A
                        DB  MCAL		; With Drive No in A Reset the Disk
			DB  ZHMDSC		; Drive to Trk 0,Sect 0
			JP  NZ,NoDisk	        ; If no disk - handle it

; Compute Size of Root
			LD DE,(MaxRootEntries)  ; Max number of root directory entires
			LD BC,32		; 32 bytes per entry
			DB MCAL		        ; Multiply DE x BC -
			DB ZIMULT		; Result in DE HL= Directory size in bytes

; HL contains the root size in bytes so div by 512 - who says the Z80 can't do fast division?
                        LD A, H                 ; Get the hi byte
                        SRL A                   ; Quick and dirty baby- just how I like it.
                        LD (RootSectors),A      ; A contains root sectors,- job done!


; Calculate location of Root Directory and start of Data area
		        LD HL,(FatSize16)       ; Get Fat sectors(low)
                	LD BC,(RsvdSectCnt)     ; Get number of reserved sectors
			ADD HL,BC		; Add reserved sectors to no of FAT sectors
			LD (RootStart),HL	; Save the starting sector of root
                        LD DE,(RootSectors)     ; Get num of root sectors
                        ADD HL,DE               ; Add root size to fat size
                        LD (DataSector),HL      ; Store start of Data area in sectors

; Load Root Directory into memory

                        LD DE,(RootSectors)     ; No of root sectors
                        LD BC,(BytesPerSector)  ; convert to bytes
                        DB MCAL		        ; Multiply DE x BC -
			DB ZIMULT		; result in HL DE= Size of root in bytes
                	LD DE,(MemLoad)         ; Memory destination for load
                        ADD HL,DE               ; Add root size to load address
                        EX DE,HL		; Swap vectors HL = mem DE = Mem+ root size
                        PUSH HL                 ; Save load address
			LD HL,(RootStart)       ; Get sector start of root
			CALL LBA2CHS            ; Convert to sector/trk in bc
                        POP HL                  ; Restore load address
                        LD A,(BootDrive)	; Get boot drive
                        DB MCAL		        ; Read block from disk, A = Drv,
			DB ZRBLK		; HL=mem start, DE=end, BC=Sct/Trk
                        JP NZ,DiskError         ; Deal with bad read

; Parse ROOT dir for cmd file. If not found it's not a boot disk

			LD BC,(MaxRootEntries)  ; Max number of root directory entires
			LD HL,(MemLoad)	        ; Start of root dir in memory
NextName:	        LD DE,CmdFileName       ; Point to start of command file name
			PUSH BC		        ; Save root entry count
			PUSH HL		        ; Save start of dir entry name in root dir
			LD B,0BH		; 11 chars in a file name
NextLetter:	        LD A,(DE)		; Move command filename letter into A
			CP (HL)		        ; Compare command filename letter with dir char
			JR NZ,NoMatch	        ; No match = not this name!
			INC DE         	        ; OK letters match so point to next command file char
			INC HL		        ; Point to dir entry next letter
			DJNZ NextLetter         ; Compare next letters
			POP HL                  ; Not needed - clear from stack
			POP BC        	        ; Not needed - clear from stack
			JR LoadFAT		; Houston we have a match! Now jump to load FAT
NoMatch:	        POP HL		        ; Get start of Dir entry name
			LD DE,32		; Size of a dir entry
			ADD HL,DE	        ; HL now points to start of next dir name
			POP BC		        ; Retreive no of root dirs
			DEC BC		        ; Reduce root dir count by 1
			LD A,B		        ; Dec rp doesn't set flags (bug) so ...
			OR C		        ; Now dec should set the carry flag properly
			JR NZ,NextName	        ; Search the next root dir name entry
			JP NoBootDisk           ; Report not a bootable disk and quit

; Find the command files starting cluster and then load FAT into memory
; (to save space FAT overwrites the ROOT dir since we no longer need it)

LoadFAT:
                        LD DE,26		; HL points to start of Name add 26 bytes
                        ADD HL,DE		; Now HL points to (byte 26) files 1st cluster
			LD E,(HL)		; Move cluster number in HL into DE
			INC HL                  ; Next byte of address
			LD D,(HL)               ; Get cluster number Hi
			LD (Cluster),DE	        ; Save the cluster number

; Calculate Size & Location of FAT

			LD DE,(FatSize16)	; Get number of FAT sectors
			LD BC,(BytesPerSector)  ; Move bytes per sector into BC
			DB MCAL		        ; Multiply DE x BC -
			DB ZIMULT		; Result in DE HL= Total FAT bytes
                        LD  DE,(MemLoad)	; Point to start of memory load
			ADD HL,DE		; Add to start address
			EX DE,HL		; Swap vectors - DE now holds mem end
                        LD BC,(RsvdSectCnt)     ; Actual no = 1st no after reserved = FAT
			LD B,C                  ; Sectors into B
			LD C,0                  ; Track is zero
                        LD A,(BootDrive)	; Drive in A
                        DB MCAL		        ; Read block from disk, A = Drv,
			DB ZRBLK		; HL=mem start, DE=end, BC=Sct/Trk

                        JR NZ,DiskError	        ; If failure

; Load Cmd image into memory  DE points to loaded code position at end of FAT

                        PUSH DE                 ; VIP address for command file start!

			LD HL,(Cluster)         ; Get cluster number

LoadIt:                 LD BC,(SectorsPerCluster); Sectors to read
                        CALL ReadSectors        ; read sector into memory
			PUSH DE                 ; save next load position

; compute next cluster
                        LD DE,(MemLoad)         ; start of FAT
                        LD HL,(Cluster)         ; cluster number
                        ADD HL,HL               ; each entry is 2 bytes long
                        ADD HL,DE               ; add to start = indexed to cluster number
                        LD E,(HL)               ; get low byte
                        INC HL                  ; point to next byte
                        LD D,(HL)               ; get high byte
                        EX DE,HL                ; cluster content into HL
                        LD DE,$FFFD             ; end of file/reserved/bad marker
                        PUSH HL                 ; Save cluster contents
                        SBC HL,DE               ; Is it an end of chain cluster?
                        POP HL                  ; Restore cluster contents
                        LD (Cluster),HL         ; Save it for later - yummy
                        POP DE                  ; Retrieve load address
                        JR C,LoadIt             ; OK it points to next cluster

; Address of command file is on Stack so treat it as a return address

Done:                   RET                     ; Jump to Command (DE was saved on stack)


;-------------------------------------------------------------------------

;*************************************************************************
; PROCEDURE ReadSectors
; reads C sectors from disk starting at LBA sectorHL into buffer at DE
;*************************************************************************

; Entry HL = LBA,
;       DE = Load address,
;        C = No of sectors to read
; Returns DE = Address + sectors per cluster


ReadSectors:            PUSH AF
                       
                        CALL Clust2LBA          ; covert cluster to sectors
                        LD A,C                  ; save sectors to read
                        CALL LBA2CHS            ; Convert to start sect/trk
                        PUSH BC                 ; save sect/track start
                        PUSH DE                 ; save load address
                        LD B,0                  ; clear hi byte
                        LD C,A                  ; get no of sectors to read
                        LD DE,(BytesPerSector)  ; multiply BC by 512 = bytes
                        DB MCAL                 ; MOS
                        DB ZIMULT               ; HL contains total bytes
                        POP DE                  ; retrieve load address
                        ADD HL,DE               ; add bytes to start address
                        EX DE,HL                ; swap vectors Hl = start DE = end
                        POP BC                  ; retrive start sect/track
                        LD A,(BootDrive)        ; Drive to read
                        DB MCAL                 ; MOS
                        DB ZRBLK                ; read block to memory
                        
                        POP AF
                        RET

;MCAL A4 ZRBLK	Read Block
;Read a block of data from the disc

;Values Passed
;Drive no. (0-3) in A register
;Start address in HL pair
;Finish address in DE pair
;Sector (0-9) in B register
;track (0-27H) in C register


;-------------------------------------------------------------------------
 ;*************************************************************************
; PROCEDURE Cluster2LBA
; convert FAT cluster in HL into LBA addressing scheme
; LBA = cluster  * sectors per cluster
;*************************************************************************
; Entry: HL = cluster number
;  Exit: HL = Offset Sector of Data



Clust2LBA:
                        PUSH BC
                        PUSH DE

                        LD D,H
                        LD E,L

                        LD BC,(SectorsPerCluster)
                        DB MCAL		        ; Multiply DE x BC -
		        DB ZIMULT		; Result in HL =
                        LD DE,(DataSector)      ; Base offset for data
                        ADD HL,DE               ; Offset into data
                        
                        POP DE
                        POP BC
                        RET


;---------------------------------------------------------------------------

;***************************************************************************
; PROCEDURE LBA2CHS
; convert HL LBA addressing scheme to CHS addressing scheme
; currently it doesn't bother with Heads - done by the FDD controller
;***************************************************************************

 ;  On entry HL contains LBA sectors
 ;  Returns BC sect/trk

LBA2CHS:                PUSH AF
                        PUSH DE
                        PUSH HL


                        LD BC,$FFFF             ; Set track counter to -1
                        LD DE,(SecPerTrk)       ; divide LBA sectors by sect/trk
ReDiv:                  SBC HL,DE               ; Divide by sectors per track
                        INC BC                  ; inc track counter
                        JR Z,Here               ; exit if divion = zero
                        JR NC,ReDiv             ; else repeat until carry

Here:                   LD (ABS_Track),BC       ; Save Tracks
                        DEFB MCAL               ; MOS
                        DEFB ZIMULT             ; Multiply tracks by sect/trk
                        POP DE                  ; retrieve LBA
                        PUSH DE
                        EX DE,HL                ; Restore total sectors
                        SBC HL,DE               ; Take away (sectors X tracks)
                        LD (ABS_Sector),HL      ; Store sector number
                        LD BC,(ABS_Track)       ; Get track num - low in c
                        LD B,L                  ; Get sector num - low in L

                        POP HL
                        POP DE                  ; Restore DE
                        POP AF                  ; Restore AF
                        RET

; ----------------------------------------------------------------
;              Error Handling and abnormal termination
;-----------------------------------------------------------------

DiskError:	        DB MCAL
			DB ZPRM
			DB "DISK I/O Error"
			DB CR,(LF+HI)
			JR KeyPressMsg

NoBootDisk:	        DB MCAL
			DB ZPRM
			DB "Not a Boot Disk!"
			DB CR,(LF+HI)
KeyPressMsg:            DB MCAL
			DB ZPRM
			DB "Press A Ke"
			DB (HI+'y')
			DB MCAL
			DB ZKEYIN
			JR EndItAll

NoDisk:		        DB MCAL
			DB ZPRM
			DB "No Disc "
			DB CR,LF
			DB "Type Drive or Enter to Quit"
                        DB (HI +':')
			DB MCAL
			DB ZKEYIN	        ; Get key press
                        CP CR		        ; = Enter?
                        JR Z, EndItAll
                        LD BC,(DriveMask)
                        SUB C
                        AND 3
                        PUSH AF
                        DB MCAL
			DB ZCRLF		; Throw a new line
                        POP AF
                        JP DriveRetry	        ; Try drive again

EndItAll:	        LD SP,(OldSP)	        ; Restore MOS stackDB	MCAL
                       	DB MCAL
                        DB ZINIT		; Drop into MOS

; -------------------------------------------------------------------------------------
;                      INTERNAL DATA SECTION
; -------------------------------------------------------------------------------------


ABS_Sector              DW 0
ABS_Track               DW 0
BootDrive:		DB "0"                  ; Can change to 'A' if you prefer
DriveMask:		DB "0"                  ; if changed to 'A' uses DOS drive char
Cluster:		DW 0
CmdFileName:            DB "XCOMMANDBIN"
DataSector:		DW 0
FATSect:		DW 0
RootStart:		DW 0
RootSectors:		DB 0
MemLoad:		DW 0300H
OldSP:			DW 0

; these are just filler bytes to the page boundary

Bottom:                 DS 29

; This is the MS boot block ID signature - I will be changing it

                        DW $AA55

; ------------------  Boot Sector Ends Here -----------------------------


; *********************************************************************
; THIS IS THE FILE ALLOCATION TABLE (F.A.T)
; This is where clusters are stored and allocated to files
; *********************************************************************

Fat1:                   DW $0001               ; Cluster 0
                        DW $FFFF               ; Cluster 1
                        DS ((512*2)-4),0       ; the rest are free

; **********************************************************************
; THIS IS THE ROOT DIRECTORY WHERE THE FILE DETAILS ARE STORED
; **********************************************************************

Root:                   DM "XCOMMANDBIN"
                        DB 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
FClust:                 DW $0000  ; points to first cluster of file (0)
                        DB 0,0,0,0
                        DS (512*8)-32,$E5

;***********************************************************************
; THIS IS THE DATA SECTION OF THE DISK - 
;  The code below is the contents of XCOMMAND.BIN

;  To use this file replace the code between SignOnMessage and EndProg
;  with whatever code you want.
;   ( I will be using it to load my new DOS for the Einy)
;  Currently the file occupies 2 clusters (4 sectors) to make it bigger -
;  link in more clusters - see FAT 1 and adjust the end space

;**********************************************************************

Data:

SignOnMsg:	        DB MCAL
			DB ZPRM 		; Show the signon logo
			DM "FX-DOS80 Ver 1.0 Boot Demo"
			DM " (C) Phil Simmons 2007"
			DB CR,LF
			DB "Press Any Ke"
			DB (HI+'y')
			DB MCAL
			DB ZKEYIN
			LD SP,(OldSP)	        ; Restore MOS stack
                        RET

EndProg:                DS (1024-(EndProg-SignOnMsg)),'D'


BootStack:              ORG $7E00
.END




